@@ -6,11 +6,10 @@
  *		as published by the Free Software Foundation; either version
  *		2 of the License, or (at your option) any later version.
  *
- * Authors:     Patrick Caulfield <patrick@debian.org)
+ * Authors:     Christine Caulfield <christine.caulfield@googlemail.com)
  *              bits based on rtmon.c by Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
  *
  */
-
 #include <sys/types.h>
 #include <linux/types.h>
 #include <stdio.h>
@@ -22,9 +21,11 @@
 #include <signal.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <sys/un.h>
 #include <sys/time.h>
 #include <netinet/in.h>
 #include <string.h>
+#include <limits.h>
 #include <linux/netfilter_decnet.h>
 #include <netdnet/dnetdb.h>
 #include <features.h>    /* for the glibc version number */
@@ -54,7 +55,8 @@
 #endif
 
 /* Where we write our status info to */
-#define STATUS_FIFO "/var/run/dnroute.status"
+#define STATUS_SOCKET "/var/run/dnroute.status"
+#define PIDFILE "/var/run/dnroute.pid"
 
 /* What we stick in the hash table */
 struct nodeinfo
@@ -85,8 +87,10 @@ extern char *if_index_to_name(int ifindex);
 extern int if_name_to_index(char *name);
 extern int send_level1_msg(struct routeinfo *);
 extern int send_level2_msg(struct routeinfo *);
+extern int pidfile_create(const char *pidFile, pid_t pid);
 
 int dnet_socket;
+static int info_socket;
 static int debugging;
 static int verbose;
 static int send_routing;
@@ -124,7 +128,7 @@ static void read_conffile(void)
 {
 	char line[255];
 
-	FILE *fp = fopen("/etc/dnroute.conf", "r");
+	FILE *fp = fopen(SYSCONF_PREFIX "/etc/dnroute.conf", "r");
 	if (!fp)
 		return;
 
@@ -192,12 +196,20 @@ static void read_conffile(void)
 /* Send network status down the FIFO */
 static void do_show_network(void)
 {
-	FILE *fp=fopen(STATUS_FIFO, "w");
+	FILE *fp;
 	int i;
 	int first = 1;
 	unsigned char dn_addr[2];
 	struct nodeent *ne;
+	struct sockaddr saddr;
+	socklen_t addrlen;
+	int new_fd;
 
+	new_fd = accept(info_socket, &saddr, &addrlen);
+	if (new_fd < 0)
+		return;
+
+	fp=fdopen(new_fd, "w");
 	if (!fp)
 	{
 		debuglog("Can't send status to FIFO\n");
@@ -214,6 +226,7 @@ static void do_show_network(void)
 				struct nodeinfo *n;
 				unsigned short area_node = area_table[i].router;
 				char *ifname = "";
+				int local = 0;
 
 				if (first)
 				{
@@ -225,6 +238,7 @@ static void do_show_network(void)
 				{
 					area_node = exec_addr->a_addr[1] << 8 | exec_addr->a_addr[0];
 					ifname = "(local)";
+					local = 1;
 				}
 
 				n = dm_hash_lookup_binary(node_hash, (void*)&area_table[i].router, 2);
@@ -237,7 +251,7 @@ static void do_show_network(void)
 					n?if_index_to_name(n->interface):ifname,
 					area_node>>10, area_node & 0x03FF,
 					ne?ne->n_name:"",
-					area_table[i].manual?"(M)":"");
+					(area_table[i].manual && !local)?"(M)":"");
 			}
 		}
 	}
@@ -330,6 +344,14 @@ static int edit_dev_route(int function, unsigned short node, int interface)
 		char                    buf[1024];
 	} req;
 
+
+	if (verbose)
+	{
+		syslog(LOG_INFO, "%sing route to %d.%d via %s\n",
+		       function == RTM_NEWROUTE?"Add":"Remov",
+		       node>>10, node&0x3FF, if_index_to_name(interface));
+	}
+
 	if (no_routes)
 		return 0;
 
@@ -364,9 +386,6 @@ static int edit_via_route(int function, unsigned short addr, unsigned short via_
 		char                    buf[1024];
 	} req;
 
-	if (no_routes)
-		return 0;
-
 	if (verbose)
 	{
 		struct dn_naddr add;
@@ -382,6 +401,9 @@ static int edit_via_route(int function, unsigned short addr, unsigned short via_
 		       addr>>10, addr&0x3FF, nodename);
 	}
 
+	if (no_routes)
+		return 0;
+
 	assert (addr>>10);
 	assert (via_node);
 
@@ -720,12 +742,15 @@ static void get_neighbours(void)
 		n->scanned = 0;
 	}
 
-	/* Send messages */
-	if (send_routing)
-		send_level1_msg(node_table);
+	if (!no_routes)
+	{
+		/* Send messages */
+		if (send_routing)
+			send_level1_msg(node_table);
 
-	if (send_level2)
-		send_level2_msg(area_table);
+		if (send_level2)
+			send_level2_msg(area_table);
+	}
 }
 
 static void add_routing_neighbour(unsigned short nodeaddr, int level, int priority, int override)
@@ -786,7 +811,7 @@ static void process_level1_message(unsigned char *buf, int len, int iface)
 		i+=2;
 		start_id = buf[i] | buf[i+1]<<8;
 		i+=2; /* Start of entries */
-		debuglog("PJC start_id = %d, num_ids = %d pos=%d\n", start_id, num_ids, i);
+		debuglog("CC start_id = %d, num_ids = %d pos=%d\n", start_id, num_ids, i);
 
 		for (num = 0; num<num_ids; num++)
 		{
@@ -875,7 +900,7 @@ static void process_level2_message(unsigned char *buf, int len, int iface)
 		start_id = buf[i] | buf[i+1]<<8;
 		i+=2; /* Start of entries */
 
-		debuglog("PJC start_id = %d, num_ids = %d\n", start_id, num_ids);
+		debuglog("CC start_id = %d, num_ids = %d\n", start_id, num_ids);
 		for (num = 0; num<num_ids; num++)
 		{
 			entry = buf[i] | buf[i+1]<<8;
@@ -890,9 +915,10 @@ static void process_level2_message(unsigned char *buf, int len, int iface)
 					return;
 
 				n = dm_hash_lookup_binary(node_hash, (void*)&nodeaddr, 2);
-				add_area_routeinfo(num+start_id,
-						   (entry&0x1FF) + cost[n->interface], ((entry&0x7E00)>>9),
-						   nodeaddr);
+				if (n)
+					add_area_routeinfo(num+start_id,
+							   (entry&0x1FF) + cost[n->interface], ((entry&0x7E00)>>9),
+							   nodeaddr);
 			}
 			i+=2;
 		}
@@ -955,6 +981,7 @@ int main(int argc, char **argv)
 	struct timespec ts;
 	int no_daemon=0;
 	mode_t oldmode;
+	struct sockaddr_un sockaddr;
 
 	/* Initialise the node hash table */
 	node_hash = dm_hash_create(1024);
@@ -1036,6 +1063,9 @@ int main(int argc, char **argv)
 			exit(0);
 		}
 
+		/* There should only be one of us */
+		pidfile_create(PIDFILE, getpid());
+
 		/* Detach ourself from the calling environment */
 		devnull = open("/dev/null", O_RDWR);
 		close(0);
@@ -1053,13 +1083,32 @@ int main(int argc, char **argv)
 	signal(SIGPIPE, SIG_IGN);
 
 	oldmode = umask(0);
-	unlink(STATUS_FIFO);
-	mkfifo(STATUS_FIFO, 0600);
+	chmod(STATUS_SOCKET, 0660);
 	umask(oldmode);
 
 	signal(SIGUSR1, usr1_sig);
 	signal(SIGALRM, alarm_sig);
 
+	/* Socket for sending "SHOW NETWORK" information */
+	unlink(STATUS_SOCKET);
+	info_socket = socket(AF_UNIX, SOCK_STREAM, PF_UNIX);
+	if (info_socket < 0)
+	{
+		syslog(LOG_ERR, "Unable to open Unix socket for information output: %m\n");
+		return 1;
+	}
+	fcntl(info_socket, F_SETFL, fcntl(info_socket, F_GETFL, 0) | O_NONBLOCK);
+
+	strcpy(sockaddr.sun_path, STATUS_SOCKET);
+	sockaddr.sun_family = AF_UNIX;
+	if (bind(info_socket, (struct sockaddr *)&sockaddr, sizeof(sockaddr)))
+	{
+		syslog(LOG_ERR, "Unable to bind Unix socket for information output: %m\n");
+		return 1;
+	}
+	/* Wait for connections */
+	listen(info_socket, 5);
+
 	/* Socket for listening for routing messages
 	   and sending our own */
 	dnet_socket = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_DNA_RT));
@@ -1069,6 +1118,8 @@ int main(int argc, char **argv)
 		return 1;
 	}
 
+	fcntl(dnet_socket, F_SETFL, fcntl(dnet_socket, F_GETFL, 0) | O_NONBLOCK);
+
 	/*
 	 * Add an entry for our area. If we are a level2 router, then it's us.
 	 */
@@ -1091,20 +1142,21 @@ int main(int argc, char **argv)
 	while (running)
 	{
 		fd_set fds;
-		unsigned char buf[1024];
+		unsigned char buf[2048];
 		sigset_t ss;
 		int len;
 		int status;
 
 		FD_ZERO(&fds);
 		FD_SET(dnet_socket, &fds);
+		FD_SET(info_socket, &fds);
 		sigfillset(&ss);
 		sigdelset(&ss, SIGUSR1);
 		sigdelset(&ss, SIGALRM);
 		sigdelset(&ss, SIGTERM);
 		sigdelset(&ss, SIGINT);
 
-		status = pselect(dnet_socket+1, &fds, NULL, NULL, NULL, &ss);
+		status = pselect(FD_SETSIZE, &fds, NULL, NULL, NULL, &ss);
 
 		if (running)
 		{
@@ -1119,21 +1171,20 @@ int main(int argc, char **argv)
 					syslog(LOG_ERR, "Error reading DECnet messages: %m\n");
 			}
 
+			if (FD_ISSET(info_socket, &fds))
+			{
+				do_show_network();
+			}
+
 			/* Things that interrupt our sleep */
 			if (alarm_rang)
 			{
 				get_neighbours();
 				alarm_rang = 0;
 			}
-
-			if (show_network)
-			{
-				do_show_network();
-				show_network = 0;
-			}
 		}
 	}
-
-	unlink(STATUS_FIFO);
+	close(dnet_socket);
+	close(info_socket);
 	exit(0);
 }
